"""Created on Tue Mar 16 22:25:59 2021.

@author: mofarrag
"""
import numpy as np

CWEIR = 1.704
BWIDTH = 10
HLIN = 0.001
power = 2 / 3


def HBC(Sub, River, Hbnd, dt, dx, inih, storewl, MinQ):

    xs = np.zeros(shape=(Sub.XSno, 8))
    Diff_coeff = np.zeros(shape=(Sub.XSno, 3))
    XSarea = np.zeros(shape=(Sub.XSno))

    wl = np.zeros(shape=(Sub.XSno))
    hx = np.zeros(shape=(Sub.XSno))
    qx = np.zeros(shape=(Sub.XSno))

    h = np.zeros(shape=(Sub.XSno, River.TS))
    q = np.zeros(shape=(Sub.XSno, River.TS))

    storewl = np.zeros(shape=(River.XSno, 2))
    # OverTopFlow = np.zeros(shape=(River.XSno*2,24))
    # OverTopWL = np.zeros(shape=(River.XSno*2,24))
    Lateral_q = MinQ.values[:-1, :-2]

    q_outL = np.zeros(shape=(Sub.XSno, River.TS))
    q_outR = np.zeros(shape=(Sub.XSno, River.TS))
    q_out = np.zeros(shape=(Sub.XSno, River.TS))

    q_outLx = np.zeros(shape=(Sub.XSno))
    q_outRx = np.zeros(shape=(Sub.XSno))
    q_outx = np.zeros(shape=(Sub.XSno))

    for t in range(0, River.TS):
        dtrest = dt
        # IC
        if t == 0:
            # fill the water depth array with the initial condition at time step 1
            hx[:] = inih
        else:
            # water level in the previous time-step is now turned to be the current
            hx = h[:, t - 1]

        # boundary condition
        # water level at the first cross section (x=1]is the BC
        hx[0] = Hbnd[0]
        # hx is the water depth of the previous time step
        # except for the first cross section water depth is the boundary condition
        # for this time step
        # hx=[t,t-1,t-1,t-1,.............]
        adaptiveTS_counter = 0
        # for each time step use adaptive time step
        # adaptive time step -----------------------------------------------------
        while dtrest > 0.001:
            adaptiveTS_counter = adaptiveTS_counter + 1
            # calculate the area and perimeter based on the water depth of the previous time step
            # for the whole cross sections in the sub-basin
            for x in range(0, Sub.XSno):
                # xs[:,:] = 0
                # calculate the area & perimeter of the whole XS
                Coords = River.getVortices(
                    hx[x],
                    Sub.hl[x],
                    Sub.hr[x],
                    Sub.cl[x],
                    Sub.cr[x],
                    Sub.mw[x],
                    Sub.Dbf[x],
                )
                xs[x, 0:2] = River.polygonGeometry(Coords)

                # area & perimeter of the Upper part only
                if hx[x] <= Sub.Dbf[x]:
                    xs[x, 2] = 0
                    xs[x, 3] = 0
                else:
                    # area of upper part only
                    xs[x, 2] = xs[x, 0] - (Sub.Dbf[x] * Sub.mw[x])
                    # perimeter of upper part
                    xs[x, 3] = xs[x, 1] - (2 * Sub.Dbf[x] + Sub.mw[x])

                # area & perimeter of the lower part only
                xs[x, 4] = xs[x, 0] - xs[x, 2]
                xs[x, 5] = xs[x, 1] - xs[x, 3]

                # to calculate hydraulic radius perimeter is in the denominator so it has not to be zero
                if xs[x, 1] == 0 or xs[x, 5] == 0:
                    assert False, (
                        "at x= "
                        + str(Sub.xsid[x])
                        + " t= "
                        + str(t)
                        + " total perimiter is zero"
                    )

            # for each cross section -------------------------------
            for x in range(0, Sub.XSno):
                # print(*,'(a3,i4,a4,i6,a]'] 't= ',t,' x= ', xsid[x],'------------------'
                # forward difference
                # surface slope is calculated based on the previous time step water depth
                # except for the first cross section hx[1]is from current time step
                # while hx[2]is from previous time step
                if x < Sub.XSno - 1:
                    # friction slope = (so - dhx/dx] - diffusive wave
                    sf = (
                        hx[x] + Sub.bedlevel[x] - hx[x + 1] - Sub.bedlevel[x + 1]
                    ) / dx
                else:
                    # for LOWER BOUNDARY node
                    sf = (
                        hx[x - 1] + Sub.bedlevel[x - 1] - hx[x] - Sub.bedlevel[x]
                    ) / dx

                # if sf <= 0 :
                # print[800000,'(a]'] '----------'
                # print(*,'(a4,i4,2x,a6,G10.4]'] 'XS= ',xsid[x], ', sf= ', sf
                # print[800000,'(a7,i4,a5,i6,2x,a6,G10.4]'] 'min = ',t ,' XS= ',xsid[x], ', sf= ', sf

                if River.D1["ModelMode"] == 2:
                    # check for the dike overtopping
                    # to LEFT?
                    # overtoping dischargee is zero unless water level exceeds the dike level
                    q_outLx[x] = 0
                    # check first the water level in the 2d grid
                    # if water level in the 1D river is less than the water depth in
                    # nighbouring cell in the 2D grid don't calculate overtopping

                    if storewl[Sub.xsid[x]] - Sub.bedlevel[x] < hx[x]:

                        # if water depth is higher than the left dike height and the threshold
                        if hx[x] > max(
                            Sub.zl[x] - Sub.bedlevel[x], River.D1["MinDepth"]
                        ):
                            # weir formular - free flow
                            q_outLx[x] = (
                                CWEIR
                                * BWIDTH
                                * (
                                    hx[x]
                                    - max(
                                        Sub.zl[x] - Sub.bedlevel[x],
                                        River.D1["MinDepth"],
                                    )
                                )
                                ** 1.5
                            )

                    # to RIGHT?
                    q_outRx[x] = 0
                    # nxs is added to the ipf just because of the numbering system of the
                    # cross section
                    if storewl[Sub.xsid[x] + River.XSno] - Sub.bedlevel[x] < hx[x]:

                        if hx[x] > max(
                            Sub.zr[x] - Sub.bedlevel[x], River.D1["MinDepth"]
                        ):
                            q_outRx[x] = (
                                CWEIR
                                * BWIDTH
                                * (
                                    hx[x]
                                    - max(
                                        Sub.zr[x] - Sub.bedlevel[x],
                                        River.D1["MinDepth"],
                                    )
                                )
                                ** 1.5
                            )

                    # correction factor
                    q_outx[x] = q_outLx[x] + q_outRx[x]
                    if q_outx[x] == 0:
                        corrL = 0
                        corrR = 0
                    else:
                        corrL = q_outLx[x] / q_outx[x]
                        corrR = q_outRx[x] / q_outx[x]

                # compute mean area and hydraulic radius
                # all cross section except the last one
                if x < Sub.XSno - 1:
                    if hx[x] <= Sub.Dbf[x]:
                        # upper discharge
                        xs[x, 6] = 0
                        # diffusive wave coefficient for upper wave
                        Diff_coeff[x, 0] = 0
                        # calculate only lower part
                        # Lower mean area and hydraulic radius
                        Area_LM = (xs[x, 4] + xs[x + 1, 4]) / 2  # mean area
                        R_LM = (
                            xs[x, 4] / xs[x, 5] + xs[x + 1, 4] / xs[x + 1, 5]
                        ) / 2  # mean radius
                    else:
                        # Upper mean area and hydraulic radius
                        Area_UM = (xs[x, 2] + xs[x + 1, 2]) / 2  # mean area
                        # print(*,'(a7,f8.3]'] 'A_UM = ', Area_UM
                        if xs[x, 3] != 0 and xs[x + 1, 3] == 0:
                            R_UM = (xs[x, 2] / xs[x, 3]) / 2  # mean radius
                        elif xs[x, 3] == 0 and xs[x + 1, 3] != 0:
                            R_UM = (xs[x + 1, 2] / xs[x + 1, 3]) / 2  # mean radius
                        elif xs[x, 3] == 0 and xs[x + 1, 3] == 0:
                            R_UM = 0  # mean radius
                        else:
                            R_UM = (
                                xs[x, 2] / xs[x, 3] + xs[x + 1, 2] / xs[x + 1, 3]
                            ) / 2  # mean radius

                        # upper discharge
                        xs[x, 6] = (1.0 / Sub.mn[x]) * Area_UM * (R_UM**power)
                        # diffusive wave coefficient for upper wave
                        Diff_coeff[x, 0] = (1.0 / Sub.mn[x]) * (R_UM**power)

                        # Lower mean area and hydraulic radius
                        Area_LM = (xs[x, 4] + xs[x + 1, 4]) / 2  # mean area
                        R_LM = (
                            xs[x, 4] / xs[x, 5] + xs[x + 1, 4] / xs[x + 1, 5]
                        ) / 2  # mean radius
                else:  # the last cross section
                    if hx[x] <= Sub.Dbf[x]:
                        # upper discharge
                        xs[x, 6] = 0
                        # diffusive wave coefficient for upper wave
                        Diff_coeff[x, 0] = 0
                        # calculate only lower part
                        # Lower mean area and hydraulic radius
                        Area_LM = (
                            xs[x, 4] + xs[x - 1, 4]
                        ) / 2  # for LOWER BOUNDARY node
                        R_LM = (
                            xs[x, 4] / xs[x, 5] + xs[x - 1, 4] / xs[x - 1, 5]
                        ) / 2  # mean radius
                    else:  # for LOWER BOUNDARY node
                        # Upper mean area and hydraulic radius
                        Area_UM = (xs[x, 2] + xs[x - 1, 2]) / 2

                        if xs[x, 3] != 0 and xs[x - 1, 3] == 0:
                            R_UM = (xs[x, 2] / xs[x, 3]) / 2  # mean radius
                        elif xs[x, 3] == 0 and xs[x - 1, 3] != 0:
                            R_UM = (xs[x - 1, 2] / xs[x - 1, 3]) / 2  # mean radius
                        elif xs[x, 3] == 0 and xs[x - 1, 3] == 0:
                            R_UM = 0
                        else:
                            R_UM = (
                                xs[x, 2] / xs[x, 3] + xs[x - 1, 2] / xs[x - 1, 3]
                            ) / 2  # mean radius

                        # upper discharge
                        xs[x, 6] = (1.0 / Sub.mn[x]) * Area_UM * (R_UM**power)
                        # diffusive wave coefficient for upper wave
                        Diff_coeff[x, 0] = (1.0 / Sub.mn[x]) * (R_UM**power)

                        # Lower mean area and hydraulic radius
                        Area_LM = (
                            xs[x, 4] + xs[x - 1, 4]
                        ) / 2  # for LOWER BOUNDARY node
                        R_LM = (
                            xs[x, 4] / xs[x, 5] + xs[x - 1, 4] / xs[x - 1, 5]
                        ) / 2  # mean radius

                # calculation of fluxes
                if abs(sf) > HLIN / dx:
                    # diffusive wave coefficient for upper wave
                    Diff_coeff[x, 0] = Diff_coeff[x, 0] / np.sqrt(abs(sf))
                    # diffusive wave coefficient for Lower wave
                    Diff_coeff[x, 1] = (
                        (1.0 / Sub.mn[x]) * (R_LM**power) / np.sqrt(abs(sf))
                    )
                    # total diffusion coefficient
                    Diff_coeff[x, 2] = Diff_coeff[x, 0] + Diff_coeff[x, 1]
                    # discharge
                    # upper discharge
                    xs[x, 6] = xs[x, 6] * (np.sqrt(abs(sf)) * sf / abs(sf))
                    # Lower discharge
                    xs[x, 7] = (
                        (1.0 / Sub.mn[x])
                        * Area_LM
                        * (R_LM**power)
                        * (np.sqrt(abs(sf)) * sf / abs(sf))
                    )
                else:
                    # diffusive wave coefficient for upper wave
                    Diff_coeff[x, 0] = Diff_coeff[x, 0] / np.sqrt(HLIN / dx)

                    # diffusive wave coefficient for Lower wave
                    Diff_coeff[x, 1] = (
                        (1.0 / Sub.mn[x]) * (R_LM**power) / np.sqrt(HLIN / dx)
                    )
                    # total diffusion coefficient
                    Diff_coeff[x, 2] = Diff_coeff[x, 0] + Diff_coeff[x, 1]

                    # discharge
                    # upper discharge
                    xs[x, 6] = xs[x, 6] * (np.sqrt(dx / HLIN) * sf)
                    # xs[x,7] = xs[x,7] * (np.sqrt(HLINQ/dx]]
                    # Lower discharge
                    xs[x, 7] = (
                        (1.0 / Sub.mn[x])
                        * Area_LM
                        * (R_LM**power)
                        * (np.sqrt(dx / HLIN) * sf)
                    )
                    # xs[x,8] = (1.0/mn[x]] * Area_LM * (R_LM** (2.0/3.0]] * (np.sqrt(HLINQ/dx]]

                # total Discharge
                qx[x] = xs[x, 6] + xs[x, 7]

                # 			if qx[x] < 0:
                # print(*,'(a,i4]'] 'Q is -ve at XS= ',xsid[x]
                # print[800000,'(a7,i4,a17,i6,a4,f8.3]'] 'min = ',t ,' Q is -ve at XS= ',xsid[x],' Q= ',qx[x]
                # print[800000,'(a,f8.3]'] 'Q is -ve at Q= ',qx[x]
                # exit

                if River.D1["ModelMode"] == 2:
                    # if the overtopping is less than 0.5 the flow
                    if qx[x] > 2 * q_outx[x]:
                        # the whole overtopping will be considered
                        qx[x] = qx[x] - q_outx[x]
                    else:
                        # if the overtopping is greater than 0.5 the flow
                        # overtopping will be 0.5 the flow only
                        q_outx[x] = 0.5 * qx[x]
                        q_outLx[x] = corrL * q_outx[x]
                        q_outRx[x] = corrR * q_outx[x]
                        qx[x] = qx[x] - q_outx[x]

            # for each cross section -------------------------------
            # print(*,'(a17,f8.3]'] 'max diff coeff = ', maxval[Diff_coeff(:,3]]

            # adjusting the time step
            dto = (dx**2) / (2 * max(Diff_coeff[:, 2]))

            if dto > dt:
                dto = dt
            else:
                if dto == 0:
                    print("adaptive time step is almost zero dto= " + str(t))
                    # arr = Diff_coeff(:,3]
                    # call Find_FInArrayF[maxval[arr],arr,loc1[1]]
                    # loc =maxloc (Diff_coeff(:,3]]
                    # print(*,'(a37,i4]'] 'At XS= ', xsid[loc1[1]]
                    # print(*,'(a37,i4]'] 'Diff coeff= ', Diff_coeff[loc1[1],3]
                    # print(*,'(a37,i4]'] 'discharge= ', qx[loc1[1]]
                    assert False, "dto= " + str(dto)

            if dto > dtrest:
                dto = dtrest

            # print[800000,'(a3,f8.2,a7,f4.2,a7,f8.2]'] 'dto= ', dto,' ratio=', dto/dt, ' max Q= ',qx[Sub.XSno]

            # update the water level at each cross section except the upstream boundary node
            for x in range(1, (Sub.XSno) - 2):
                # check if the XS has laterals
                # FindInArrayF(real[xsid[x]],sub_XSLaterals,loc)
                if Sub.xsid[x] in Sub.LateralsTable:
                    loc = Sub.LateralsTable.index(Sub.xsid[x])
                    # have only the laterals
                    qlat = Lateral_q[t, loc]
                    # print(*,*] "Lat= ", qlat
                else:
                    qlat = 0

                # Forward time backward scpace
                XSarea[x] = (
                    xs[x, 0] - ((qx[x] - qx[x - 1]) * (dto / dx)) + qlat * (dto / dx)
                )  # dto # difference between two consecutive XS areas

                if XSarea[x] <= 0:
                    break
                    print(x)
                    print(XSarea[x])
                    assert False, "Error cross section area is less than zero"
                else:
                    # calculate the bankful area
                    bankful_area = Sub.Dbf[x] * Sub.mw[x]

                    # if the new area is less than the area of the min[hl,hr]
                    if XSarea[x] <= Sub.AreaPerLow[x, 0]:
                        # if the area is less than the bankful area
                        if XSarea[x] <= bankful_area:
                            hx[x] = XSarea[x] / Sub.mw[x]
                        else:
                            # Case 1
                            dummyarea = XSarea[x] - bankful_area
                            dummyh = QuadraticEqn(
                                Sub.cr[x] / Sub.hr[x] + Sub.cl[x] / Sub.hl[x],
                                2 * Sub.mw[x],
                                -2 * dummyarea,
                            )
                            hx[x] = dummyh + Sub.Dbf[x]

                    # if the new area is less than the area of the max[hl,hr]
                    elif XSarea[x] <= Sub.AreaPerHigh[x, 1]:
                        dummyarea = XSarea[x] - bankful_area
                        # check which dike is higher
                        if Sub.hl[x] < Sub.hr[x]:
                            # case 2
                            dummyh = QuadraticEqn(
                                Sub.cr[x] / Sub.hr[x] * 0.5,
                                Sub.mw[x] + Sub.cl[x],
                                -dummyarea - Sub.hl[x] * Sub.cl[x] * 0.5,
                            )
                        else:
                            # case 3
                            dummyh = QuadraticEqn(
                                Sub.cl[x] / Sub.hl[x] * 0.5,
                                Sub.mw[x] + Sub.cr[x],
                                -dummyarea - Sub.hr[x] * Sub.cr[x] * 0.5,
                            )

                        hx[x] = dummyh + Sub.Dbf[x]

                    else:
                        dummyarea = XSarea[x] - bankful_area
                        # case 4 (linear function]
                        dummyh = (
                            dummyarea
                            + Sub.cl[x] * Sub.hl[x] / 2.0
                            + Sub.cr[x] * Sub.hr[x] / 2.0
                        ) / (Sub.cr[x] + Sub.cl[x] + Sub.mw[x])
                        hx[x] = dummyh + Sub.Dbf[x]

                    # to prevent -ve Sf in the next time step calculation
                    if hx[x] + Sub.bedlevel[x] > hx[x - 1] + Sub.bedlevel[x - 1]:
                        # print[800000,'(a]'] '-----'
                        # print[800000,'(a,i5,a11,i2]'] 'time = ',t , ' round no= ',adaptiveTS_counter
                        # print[800000,'(a9,i5,a4,f10.3]'] 'Q at x = ', xsid[x],' Q= ' ,qx[x]
                        # print[800000,'(a9,i5,a4,f10.3]'] 'Q at x = ', xsid[x-1],' Q= ' ,qx[x-1]
                        # g = (hx[x]+Sub.bedlevel[x])- (hx[x-1]+Sub.bedlevel[x-1])
                        # print[800000,'(a,i5,a3,f10.4,a2]'] 'H is higher than previous XS= ',xsid[x], "by ", g,' m'
                        # print(*,'(a,i5]'] 'H is higher than previous XS= ',xsid[x]
                        hx[x] = hx[x - 1] + Sub.bedlevel[x - 1] - Sub.bedlevel[x] - 0.01

                    # this check to prevent the previous step from creating -ve values
                    # as after calculating the h[x]by solving the quadratic equation (QuadraticEqn]the step
                    # of hx[x] = hx[x-1] + bedlevel[x-1] - bedlevel[x] -0.01 can result in -ve hx
                    # so this check should be after not before
                    if hx[x] <= 0.01:
                        hx[x] = 0.01

            # water depth at the last point equals to wl at point before the last
            hx[Sub.XSno - 1] = hx[Sub.XSno - 2] - 0.001

            if min(hx) < 0:
                assert False, "calculated water depth is less than 0 "

            # update the discharge array & dt
            dtrest = dtrest - dto
            q[:, t] = q[:, t] + qx[:] * (dto / dt)

            if River.D1["ModelMode"] == 2:
                q_outL[:, t] = q_outL[:, t] + q_outLx * (dto / dt)
                q_outR[:, t] = q_outR[:, t] + q_outRx * (dto / dt)
                q_out[:, t] = q_outL[:, t] + q_outR[:, t]

        # adaptive time step ------------------------------------------------

        # store the calculated water level & water depth
        h[:, t] = hx
        wl = hx + Sub.bedlevel

    return q, h, wl


def QuadraticEqn(a, b, c):
    delta = (b**2) - 4 * a * c
    return (-b + np.sqrt(delta)) / (2 * a)



"""RUN.

RUN contains functions to to connect the parameter spatial distribution
function with the with both component of the spatial representation of the hydrological
process (conceptual model & spatial routing) to calculate the predicted
runoff at known locations based on given performance function

Created on Sun Jun 24 21:02:34 2018

@author: Mostafa
"""
import numpy as np
import pandas as pd
from loguru import logger

from Hapi.catchment import Catchment
from Hapi.hm.saintvenant import SaintVenant
from Hapi.rrm.wrapper import Wrapper


class Run(Catchment):
    """Run.

    Run sub-class validate the spatial data and hand it to the wrapper class, It is
    a sub-class from the catchment class, so you need to create the Catchment
    object first to run the model

    Methods:
        1- RunHapi
        2- runHAPIwithLake
        3- runFW1
        4- RunFW1withLake
        5- runLumped
    """

    def __init__(self):
        self.Qsim = None
        pass

    def RunHapi(self):
        """RunModel.

        this function runs the conceptual distributed hydrological model

        Inputs:
        ----------
        1-Paths:

        4-FlowAccPath:

        5-FlowDPath:
            [String] path to the Flow Direction raster of the catchment (it should
            include the raster name and extension)
        7-ParPath:
            [String] path to the Folder contains parameters rasters of the catchment
        8-p2:
            [List] list of unoptimized parameters
            p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
            p2[1] = catchment area in km2

        Outputs:
        ----------
            1-statevariables: [numpy attribute]
                4D array (rows,cols,time,states) states are [sp,wc,sm,uz,lv]
            2-qlz: [numpy attribute]
                3D array of the lower zone discharge
            3-quz: [numpy attribute]
                3D array of the upper zone discharge
            4-qout: [numpy attribute]
                1D timeseries of discharge at the outlet of the catchment
                of unit m3/sec
            5-quz_routed: [numpy attribute]
                3D array of the upper zone discharge  accumulated and
                routed at each time step
            6-qlz_translated: [numpy attribute]
                3D array of the lower zone discharge translated at each time step

        Example:
        ----------
            PrecPath = prec_path="meteodata/4000/calib/prec"
            Evap_Path = evap_path="meteodata/4000/calib/evap"
            TempPath = temp_path="meteodata/4000/calib/temp"
            DemPath = "GIS/4000/dem4000.tif"
            FlowAccPath = "GIS/4000/acc4000.tif"
            FlowDPath = "GIS/4000/fd4000.tif"
            ParPath = "meteodata/4000/parameters"
            p2=[1, 227.31]
            st, q_out, q_uz_routed = RunModel(PrecPath,Evap_Path,TempPath,DemPath,
                                              FlowAccPath,FlowDPath,ParPath,p2)
        """
        ### input data validation
        # data type
        # assert type(self.FlowAcc)==gdal.Dataset, "flow_acc should be read using gdal (gdal dataset please read it using gdal library) "
        # assert type(self.FlowDir)==gdal.Dataset, "flow_direct should be read using gdal (gdal dataset please read it using gdal library) "

        # input dimensions
        [fd_rows, fd_cols] = self.FlowDirArr.shape
        assert (
            fd_rows == self.rows and fd_cols == self.cols
        ), "all input data should have the same number of rows"

        # input dimensions
        assert (
            np.shape(self.Prec)[0] == self.rows
            and np.shape(self.ET)[0] == self.rows
            and np.shape(self.Temp)[0] == self.rows
            and np.shape(self.Parameters)[0] == self.rows
        ), "all input data should have the same number of rows"
        assert (
            np.shape(self.Prec)[1] == self.cols
            and np.shape(self.ET)[1] == self.cols
            and np.shape(self.Temp)[1] == self.cols
            and np.shape(self.Parameters)[1] == self.cols
        ), "all input data should have the same number of columns"
        assert (
            np.shape(self.Prec)[2] == np.shape(self.ET)[2] and np.shape(self.Temp)[2]
        ), "all meteorological input data should have the same length"

        # run the model
        Wrapper.RRMModel(self)

        print("Model Run has finished")

    def RunFloodModel(self):
        """RunFloodModel.

        this function runs the conceptual distributed hydrological model

        Inputs:
        ----------
        1-Paths:

        4-FlowAccPath:

        5-FlowDPath:
            [String] path to the Flow Direction raster of the catchment (it should
            include the raster name and extension)
        7-ParPath:
            [String] path to the Folder contains parameters rasters of the catchment
        8-p2:
            [List] list of unoptimized parameters
            p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
            p2[1] = catchment area in km2

        Outputs:
        ----------
            1-statevariables: [numpy attribute]
                4D array (rows,cols,time,states) states are [sp,wc,sm,uz,lv]
            2-qlz: [numpy attribute]
                3D array of the lower zone discharge
            3-quz: [numpy attribute]
                3D array of the upper zone discharge
            4-qout: [numpy attribute]
                1D timeseries of discharge at the outlet of the catchment
                of unit m3/sec
            5-quz_routed: [numpy attribute]
                3D array of the upper zone discharge  accumulated and
                routed at each time step
            6-qlz_translated: [numpy attribute]
                3D array of the lower zone discharge translated at each time step

        Example:
        ----------
            PrecPath = prec_path="meteodata/4000/calib/prec"
            Evap_Path = evap_path="meteodata/4000/calib/evap"
            TempPath = temp_path="meteodata/4000/calib/temp"
            DemPath = "GIS/4000/dem4000.tif"
            FlowAccPath = "GIS/4000/acc4000.tif"
            FlowDPath = "GIS/4000/fd4000.tif"
            ParPath = "meteodata/4000/parameters"
            p2=[1, 227.31]
            st, q_out, q_uz_routed = RunModel(PrecPath,Evap_Path,TempPath,DemPath,
                                              FlowAccPath,FlowDPath,ParPath,p2)
        """
        ### input data validation
        # data type
        # assert type(self.FlowAcc)==gdal.Dataset, "flow_acc should be read using gdal (gdal dataset please read it using gdal library) "
        # assert type(self.FlowDir)==gdal.Dataset, "flow_direct should be read using gdal (gdal dataset please read it using gdal library) "

        # input dimensions
        [fd_rows, fd_cols] = self.FlowDirArr.shape
        assert (
            fd_rows == self.rows and fd_cols == self.cols
        ), "all input data should have the same number of rows"

        # input dimensions
        assert (
            np.shape(self.Prec)[0] == self.rows
            and np.shape(self.ET)[0] == self.rows
            and np.shape(self.Temp)[0] == self.rows
            and np.shape(self.Parameters)[0] == self.rows
        ), "all input data should have the same number of rows"
        assert (
            np.shape(self.Prec)[1] == self.cols
            and np.shape(self.ET)[1] == self.cols
            and np.shape(self.Temp)[1] == self.cols
            and np.shape(self.Parameters)[1] == self.cols
        ), "all input data should have the same number of columns"
        assert (
            np.shape(self.Prec)[2] == np.shape(self.ET)[2] and np.shape(self.Temp)[2]
        ), "all meteorological input data should have the same length"

        assert (
            np.shape(self.BankfullDepth)[0] == self.rows
            and np.shape(self.RiverWidth)[0] == self.rows
            and np.shape(self.RiverRoughness)[0] == self.rows
            and np.shape(self.FloodPlainRoughness)[0] == self.rows
        ), "all input data should have the same number of rows"
        assert (
            np.shape(self.BankfullDepth)[1] == self.cols
            and np.shape(self.RiverWidth)[1] == self.cols
            and np.shape(self.RiverRoughness)[1] == self.cols
            and np.shape(self.FloodPlainRoughness)[1] == self.cols
        ), "all input data should have the same number of columns"

        # run the model
        Wrapper.RRMModel(self)
        print("RRM has finished")
        SV = SaintVenant()
        SV.KinematicRaster(self)
        print("1D model Run has finished")

    def runHAPIwithLake(self, Lake):
        """RunDistwithLake.

        this function runs the conceptual distributed hydrological model

        Inputs:
        ----------
        1-Paths:
            1-PrecPath:
                [String] path to the Folder contains precipitation rasters
            2-Evap_Path:
                [String] path to the Folder contains Evapotranspiration rasters
            3-TempPath:
                [String] path to the Folder contains Temperature rasters
            4-FlowAccPath:
                [String] path to the Flow Accumulation raster of the catchment (it should
                include the raster name and extension)
            5-FlowDPath:
                [String] path to the Flow Direction raster of the catchment (it should
                include the raster name and extension)
        7-ParPath:
            [String] path to the Folder contains parameters rasters of the catchment
        8-p2:
            [List] list of unoptimized parameters
            p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
            p2[1] = catchment area in km2

        Outputs:
        ----------
        1- st:
            [4D array] state variables
        2- q_out:
            [1D array] calculated Discharge at the outlet of the catchment
        3- q_uz:
            [3D array] Distributed discharge for each cell

        Example:
        ----------
            PrecPath = prec_path="meteodata/4000/calib/prec"
            Evap_Path = evap_path="meteodata/4000/calib/evap"
            TempPath = temp_path="meteodata/4000/calib/temp"
            DemPath = "GIS/4000/dem4000.tif"
            FlowAccPath = "GIS/4000/acc4000.tif"
            FlowDPath = "GIS/4000/fd4000.tif"
            ParPath = "meteodata/4000/parameters"
            p2=[1, 227.31]
            st, q_out, q_uz_routed = RunModel(PrecPath,Evap_Path,TempPath,DemPath,
                                              FlowAccPath,FlowDPath,ParPath,p2)
        """
        # input dimensions
        [fd_rows, fd_cols] = self.FlowDirArr.shape
        assert (
            fd_rows == self.rows and fd_cols == self.cols
        ), "all input data should have the same number of rows and columns"

        # input dimensions
        assert (
            np.shape(self.Prec)[0] == self.rows
            and np.shape(self.ET)[0] == self.rows
            and np.shape(self.Temp)[0] == self.rows
            and np.shape(self.Parameters)[0] == self.rows
        ), "all input data should have the same number of rows"
        assert (
            np.shape(self.Prec)[1] == self.cols
            and np.shape(self.ET)[1] == self.cols
            and np.shape(self.Temp)[1] == self.cols
            and np.shape(self.Parameters)[1] == self.cols
        ), "all input data should have the same number of columns"
        assert (
            np.shape(self.Prec)[2] == np.shape(self.ET)[2] and np.shape(self.Temp)[2]
        ), "all meteorological input data should have the same length"

        assert (
            np.shape(Lake.MeteoData)[0] == np.shape(self.Prec)[2]
        ), "Lake meteorological data has to have the same length as the distributed raster data"
        assert (
            np.shape(Lake.MeteoData)[1] >= 3
        ), "Lake Meteo data has to have at least three columns rain, ET, and Temp"

        # run the model
        Wrapper.RRMWithlake(self, Lake)

        print("Model Run has finished")

    def runFW1(self):
        """RunDistwithLake.

        this function runs the conceptual distributed hydrological model

        Inputs:
        ----------
        1-Paths:
            1-PrecPath:
                [String] path to the Folder contains precipitation rasters
            2-Evap_Path:
                [String] path to the Folder contains Evapotranspiration rasters
            3-TempPath:
                [String] path to the Folder contains Temperature rasters
            4-FlowAccPath:
                [String] path to the Flow Accumulation raster of the catchment (it should
                include the raster name and extension)
            5-FlowDPath:
                [String] path to the Flow Direction raster of the catchment (it should
                include the raster name and extension)
        7-ParPath:
            [String] path to the Folder contains parameters rasters of the catchment
        8-p2:
            [List] list of unoptimized parameters
            p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
            p2[1] = catchment area in km2

        Outputs:
        ----------
        1- st:
            [4D array] state variables
        2- q_out:
            [1D array] calculated Discharge at the outlet of the catchment
        3- q_uz:
            [3D array] Distributed discharge for each cell

        Example:
        ----------
            PrecPath = prec_path="meteodata/4000/calib/prec"
            Evap_Path = evap_path="meteodata/4000/calib/evap"
            TempPath = temp_path="meteodata/4000/calib/temp"
            DemPath = "GIS/4000/dem4000.tif"
            FlowAccPath = "GIS/4000/acc4000.tif"
            FlowDPath = "GIS/4000/fd4000.tif"
            ParPath = "meteodata/4000/parameters"
            p2=[1, 227.31]
            st, q_out, q_uz_routed = RunModel(PrecPath,Evap_Path,TempPath,DemPath,
                                              FlowAccPath,FlowDPath,ParPath,p2)
        """
        # input data validation

        # input dimensions
        assert (
            np.shape(self.Prec)[0] == self.rows
            and np.shape(self.ET)[0] == self.rows
            and np.shape(self.Temp)[0] == self.rows
            and np.shape(self.Parameters)[0] == self.rows
        ), "all input data should have the same number of rows"
        assert (
            np.shape(self.Prec)[1] == self.cols
            and np.shape(self.ET)[1] == self.cols
            and np.shape(self.Temp)[1] == self.cols
            and np.shape(self.Parameters)[1] == self.cols
        ), "all input data should have the same number of columns"
        assert (
            np.shape(self.Prec)[2] == np.shape(self.ET)[2] and np.shape(self.Temp)[2]
        ), "all meteorological input data should have the same length"

        # run the model
        Wrapper.FW1(self)

        print("Model Run has finished")

    def RunFW1withLake(self, Lake):
        """RunDistwithLake.

        this function runs the conceptual distributed hydrological model

        Parameters
        ----------
        1-Paths:
            1-PrecPath:
                [String] path to the Folder contains precipitation rasters
            2-Evap_Path:
                [String] path to the Folder contains Evapotranspiration rasters
            3-TempPath:
                [String] path to the Folder contains Temperature rasters
            4-FlowAccPath:
                [String] path to the Flow Accumulation raster of the catchment (it should
                include the raster name and extension)
            5-FlowDPath:
                [String] path to the Flow Direction raster of the catchment (it should
                include the raster name and extension)
        7-ParPath:
            [String] path to the Folder contains parameters rasters of the catchment
        8-p2:
            [List] list of unoptimized parameters
            p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
            p2[1] = catchment area in km2

        Returns
        -------
        st: [4D array]
            state variables
        q_out: [1D array]
            calculated Discharge at the outlet of the catchment
        q_uz: [3D array]
            Distributed discharge for each cell

        Example
        -------
        PrecPath = prec_path="meteodata/4000/calib/prec"
        Evap_Path = evap_path="meteodata/4000/calib/evap"
        TempPath = temp_path="meteodata/4000/calib/temp"
        DemPath = "GIS/4000/dem4000.tif"
        FlowAccPath = "GIS/4000/acc4000.tif"
        FlowDPath = "GIS/4000/fd4000.tif"
        ParPath = "meteodata/4000/parameters"
        p2=[1, 227.31]
        st, q_out, q_uz_routed = RunModel(PrecPath,Evap_Path,TempPath,DemPath,
                                          FlowAccPath,FlowDPath,ParPath,p2)
        """
        # input data validation

        # input dimensions
        assert (
            np.shape(self.Prec)[0] == self.rows
            and np.shape(self.ET)[0] == self.rows
            and np.shape(self.Temp)[0] == self.rows
            and np.shape(self.Parameters)[0] == self.rows
        ), "all input data should have the same number of rows"
        assert (
            np.shape(self.Prec)[1] == self.cols
            and np.shape(self.ET)[1] == self.cols
            and np.shape(self.Temp)[1] == self.cols
            and np.shape(self.Parameters)[1] == self.cols
        ), "all input data should have the same number of columns"
        assert (
            np.shape(self.Prec)[2] == np.shape(self.ET)[2] and np.shape(self.Temp)[2]
        ), "all meteorological input data should have the same length"

        assert (
            np.shape(Lake.MeteoData)[0] == np.shape(self.Prec)[2]
        ), "Lake meteorological data has to have the same length as the distributed raster data"
        assert (
            np.shape(Lake.MeteoData)[1] >= 3
        ), "Lake Meteo data has to have at least three columns rain, ET, and Temp"

        # run the model
        Wrapper.FW1Withlake(self, Lake)

    def runLumped(
        self,
        Route: int = 0,
        RoutingFn=None,
    ):
        """runLumped.

        this function runs lumped conceptual model

        Parameters
        ----------
        ConceptualModel: [function]
            conceptual model and it should contain a function called simulate
        data: [numpy array]
            meteorological data as array with the first column as precipitation
            second as evapotranspiration, third as temperature and forth column as
            long term average temperature
        parameters: [numpy array]
            conceptual model parameters as array
        p2: [List]
            list of unoptimized parameters
            p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
            p2[1] = catchment area in km2
        init_st: [list]
            initial state variables values [sp, sm, uz, lz, wc].
        Routing: [0 or 1]
            to decide wether t route the generated discharge hydrograph or not
        RoutingFn: [function]
            function to route the dischrge hydrograph.

        Returns
        ----------
        1- st:
            [numpy array] 3d array of the 5 state variable data for each cell
        2- q_lz:
            [numpy array] 1d array of the calculated discharge.

        examples:
        ----------
            p2=[24, 1530]
            #[sp,sm,uz,lz,wc]
            init_st=[0,5,5,5,0]
            snow=0
        """
        if RoutingFn is None:
            RoutingFn = []
        if self.TemporalResolution.lower() == "daily":
            ind = pd.date_range(self.start, self.end, freq="D")
        else:
            ind = pd.date_range(self.startdate, self.enddate, freq="H")

        Qsim = pd.DataFrame(index=ind)

        Wrapper.Lumped(self, Route, RoutingFn)
        Qsim["q"] = self.Qsim
        self.Qsim = Qsim[:]
        logger.info("Lumped model run has finished successfully")


if __name__ == "__main__":
    print("Run")



"""Console script for HapiSM."""
import argparse
import sys


def main():
    """Console script for HapiSM."""
    parser = argparse.ArgumentParser()
    parser.add_argument("_", nargs="*")
    args = parser.parse_args()

    print("Arguments: " + str(args._))
    print("Replace this message by putting your code into " "Hapi.cli.main")
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""Lumped Conceptual HBV model.

HBV is lumped conceptual model consists of precipitation, snow melt,
soil moisture and response subroutine to convert precipitation into o runoff,
where state variables are updated each time step to represent a specific
hydrologic behaviour the catchment

This version was edited based on a Master Thesis on "Spatio-temporal simulation
of catchment response based on dynamic weighting of hydrological models" on april 2018

- Main Model inputs are Precipitation, evapotranspiration and temperature,
  conversion vector P2 = [time_conversion, catchment_area], parameters.
- Optional model inputs are: initial state variables, initial discharge
  Long term average temptearature .

- Model output is Qalculated dicharge at time t+1
- Model equations are solved using explicit scheme


    Inputs
    ----------
    precipitation : array_like [n]
        Average precipitation [mm/h]
    evapotranspitration : array_like [n]
        Potential Evapotranspiration [mm/h]
    temperature : array_like [n]
        Average temperature [C]
    init_st : array_like [5], optional
        Initial model states, [sp, sm, uz, lz, wc]. If unspecified,
        [0.0, 30.0, 30.0, 30.0, 0.0] mm
    q_init : float, optional
        Initial discharge value. If unspecified set to 10.0
    ll_temp : array_like [n], optional
        Long term average temptearature. If unspecified, calculated from temp.
    p2 : array_like [2]
        Problem parameter vector setup as:
        [tfac, area]
    par : array_like [10]
        Parameter vector, set up as:
        [fc, beta, e_corr, etf, lp, c_flux, k, k1, alpha, perc]
    Returns
    -------
    q_sim : array_like [n]
        Discharge for the n time steps of the precipitation vector [m3/s]
    st : array_like [n, 5]
        Model states for the complete time series [mm]

- model structure uses 18 parameters if the catchment has snow
    [ltt, utt, rfcf, sfcf, ttm, cfmax, cwh, cfr, fc, beta, e_corr, etf, lp,
    c_flux, k, k1, alpha, perc]

    otherwise it uses 10 parameters
    [rfcf, fc, beta, etf, lp, c_flux, k, k1, alpha, perc]


    Parameters
    ----------
    ltt : float
        Lower temperature treshold [C]
    utt : float
        Upper temperature treshold [C]
    rfcf : float
        Rainfall corrector factor
    sfcf : float
        Snowfall corrector factor
    ttm : float
        Temperature treshold for Melting [C]
    cfmax : float
        Day degree factor
    cwh : float
        Capacity for water holding in snow pack
    cfr : float
        Refreezing factor
    fc : float
        Filed capacity
    beta : float
        Shape coefficient for effective precipitation separation
    e_corr : float
        Evapotranspiration corrector factor
    etf : float
        Total potential evapotranspiration
    lp : float _soil
        wilting point
    c_flux : float
        Capilar flux in the root zone
    k : float
        Upper zone recession coefficient
        Upper zone response coefficient
    k1 : float
        Lower zone recession coefficient
        Lowe zone response coefficient
    alpha : float
        Response box parameter
        upper zone runoff coefficient
    perc: float
        percolation
"""

import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline as interp11

# import sklearn.metrics as error

# HBV base model parameters
P_LB = [
    -1.5,  # ltt
    0.001,  # utt
    0.001,  # ttm
    0.04,  # cfmax [mm c^-1 h^-1]
    50.0,  # fc
    0.6,  # ecorr
    0.001,  # etf
    0.2,  # lp
    0.00042,  # k [h^-1] upper zone
    0.0000042,  # k1 lower zone
    0.001,  # alpha
    1.0,  # beta
    0.001,  # cwh
    0.01,  # cfr
    0.0,  # c_flux
    0.001,  # perc mm/h
    0.6,  # rfcf
    0.4,  # sfcf
    1,
]  # Maxbas

P_UB = [
    2.5,  # ttm
    3.0,  # utt
    2.0,  # ttm
    0.4,  # cfmax [mm c^-1 h^-1]
    500.0,  # fc
    1.4,  # ecorr
    5.0,  # etf
    0.5,  # lp
    0.0167,  # k upper zone
    0.00062,  # k1 lower zone
    1.0,  # alpha
    6.0,  # beta
    0.1,  # cwh
    1.0,  # cfr
    0.08,  # c_flux - 2mm/day
    0.125,  # perc mm/hr
    1.4,  # rfcf
    1.4,  # sfcf
    10,
]  # maxbas

DEF_ST = [0.0, 10.0, 10.0, 10.0, 0.0]
DEF_q0 = 0


def Precipitation(temp, ltt, utt, prec, rfcf, sfcf, tfac):
    """Precipitation Precipitaiton routine of the HBV96 model.

    If temperature is lower than ltt, all the precipitation is considered as
    snow. If the temperature is higher than utt, all the precipitation is
    considered as rainfall. In case that the temperature is between ltt and
    utt, precipitation is a linear mix of rainfall and snowfall.

    Parameters
    ----------
    temp : float
        Measured temperature [C]
    ltt : float
        Lower temperature treshold [C]
    utt : float
        Upper temperature treshold [C]
    prec : float
        Precipitation [mm]
    rfcf : float
        Rainfall corrector factor
    sfcf : float
        Snowfall corrector factor

    Returns
    -------
    _rf : float
        Rainfall [mm]
    _sf : float
        Snowfall [mm]
    """

    if temp <= ltt:  # if temp <= lower temp threshold
        rf = 0.0  # no rainfall all the precipitation will convert into snowfall
        sf = prec * sfcf

    elif temp >= utt:  # if temp > upper threshold
        rf = prec * rfcf  # no snowfall all the precipitation becomes rainfall
        sf = 0.0

    else:  # if  ltt< temp < utt
        rf = ((temp - ltt) / (utt - ltt)) * prec * rfcf
        sf = (1.0 - ((temp - ltt) / (utt - ltt))) * prec * sfcf

    return rf, sf


def Snow(cfmax, tfac, temp, ttm, cfr, cwh, rf, sf, wc_old, sp_old):
    """Snow Snow routine of the HBV-96 model.

    The snow pack consists of two states: Water Content (wc) and Snow Pack
    (sp). The first corresponds to the liquid part of the water in the snow,
    while the latter corresponds to the solid part. If the temperature is
    higher than the melting point, the snow pack will melt and the solid snow
    will become liquid. In the opposite case, the liquid part of the snow will
    refreeze, and turn into solid. The water that cannot be stored by the solid
    part of the snow pack will drain into the soil as part of infiltration.

    Parameters
    ----------
    cfmax : float
        Day degree factor
    tfac : float
        Temperature correction factor
    temp : float
        Temperature [C]
    ttm : float
        Temperature treshold for Melting [C]
    cfr : float
        Refreezing factor
    cwh : float
        Capacity for water holding in snow pack
    _rf : float
        Rainfall [mm]
    _sf : float
        Snowfall [mm]
    wc_old : float
        Water content in previous state [mm]
    sp_old : float
        snow pack in previous state [mm]

    Returns
    -------
    _in : float
        Infiltration [mm]
    _wc_new : float
        Water content in posterior state [mm]
    _sp_new : float
        Snowpack in posterior state [mm]
    """

    if temp > ttm:  # if temp > melting threshold
        # then either some snow will melt or the entire snow will melt
        if (
            cfmax * (temp - ttm) < sp_old + sf
        ):  # if amount of melted snow < the entire existing snow (previous amount+new)
            melt = cfmax * (temp - ttm)
        else:  # if amount of melted snow > the entire existing snow (previous amount+new)
            melt = (
                sp_old + sf
            )  # then the entire existing snow will melt (old snow pack + the current snowfall)

        sp_new = sp_old + sf - melt
        wc_int = wc_old + melt + rf

    else:  # if temp < melting threshold
        # then either some water will freeze or all the water willfreeze
        if (
            cfr * cfmax * (ttm - temp) < wc_old + rf
        ):  # then either some water will freeze or all the water willfreeze
            refr = (
                cfr * cfmax * (ttm - temp)
            )  # cfmax*(ttm-temp) is the rate of melting of snow while cfr*cfmax*(ttm-temp) is the rate of freeze of melted water  (rate of freezing > rate of melting)
        else:  # if the amount of frozen water > entire water available
            refr = wc_old + rf

        sp_new = sp_old + sf + refr
        wc_int = wc_old - refr + rf

    if wc_int > cwh * sp_new:  # if water content > holding water capacity of the snow
        inf = wc_int - cwh * sp_new  # water content  will infiltrate
        wc_new = cwh * sp_new  # and the capacity of snow of holding water will retained
    else:  # if water content < holding water capacity of the snow
        inf = 0.0  # no infiltration
        wc_new = wc_int

    return inf, wc_new, sp_new


def Soil(fc, beta, etf, temp, tm, e_corr, lp, tfac, c_flux, inf, ep, sm_old, uz_old):
    """Soil Soil routine of the HBV-96 model.

    The model checks for the amount of water that can infiltrate the soil,
    coming from the liquid precipitation and the snow pack melting. A part of
    the water will be stored as soil moisture, while other will become runoff,
    and routed to the upper zone tank.

    Parameters
    ----------
    fc : float
        Filed capacity
    beta : float
        Shape coefficient for effective precipitation separation
    etf : float
        Total potential evapotranspiration
    temp : float
        Temperature
    tm : float
        Average long term temperature
    e_corr : float
        Evapotranspiration corrector factor
    lp : float _soil
        wilting point
    tfac : float
        Time conversion factor
    c_flux : float
        Capilar flux in the root zone
    _in : float
        actual infiltration
    ep : float
        actual evapotranspiration
    sm_old : float
        Previous soil moisture value
    uz_old : float
        Previous Upper zone value

    Returns
    -------
    sm_new : float
        New value of soil moisture
    uz_int_1 : float
        New value of direct runoff into upper zone
    """

    qdr = max(
        sm_old + inf - fc, 0
    )  # direct run off as soil moisture exceeded the field capacity

    inf = inf - qdr
    r = ((sm_old / fc) ** beta) * inf  # recharge to the upper zone

    ep_int = (
        (1.0 + etf * (temp - tm)) * e_corr * ep
    )  # Adjusted potential evapotranspiration

    ea = min(ep_int, (sm_old / (lp * fc)) * ep_int)

    cf = c_flux * ((fc - sm_old) / fc)  # capilary rise

    # if capilary rise is more than what is available take all the available and leave it empty

    if uz_old + r < cf:
        cf = uz_old + r
        uz_int_1 = 0
    else:
        #        uz_int_1 = uz_old + _r - _cf
        uz_int_1 = uz_old + r - cf + qdr

    sm_new = max(sm_old + inf - r + cf - ea, 0)
    #    uz_int_1 = uz_old + _r - _cf + qdr
    return sm_new, uz_int_1


def Response(tfac, perc, alpha, k, k1, area, lz_old, uz_int_1):
    """Response The response routine of the HBV-96 model.

    The response routine is in charge of transforming the current values of
    upper and lower zone into discharge. This routine also controls the
    recharge of the lower zone tank (baseflow). The transformation of units
    also occurs in this point.

    Parameters
    ----------
    tfac : float
        Number of hours in the time step
    perc : float
        Percolation value [mm\hr]
    alpha : float
        Response box parameter
        upper zone runoff coefficient
    k : float
        Upper zone recession coefficient
        Upper zone response coefficient
    k1 : float
        Lower zone recession coefficient
        Lowe zone response coefficient
    area : float
        Catchment area [Km2]
    lz_old : float
        Previous lower zone value [mm]
    uz_int_1 : float
        Previous upper zone value before percolation [mm]
    qdr : float
        Direct runoff [mm]
    """
    # upper zone
    # if perc > Quz then perc = Quz and Quz = 0 if not perc = value and Quz= Quz-perc so take the min
    uz_int_2 = np.max([uz_int_1 - perc, 0.0])  # upper zone after percolation
    _q_0 = k * (uz_int_2 ** (1.0 + alpha))

    if _q_0 > uz_int_2:  # if q_0 =30 and UZ=20
        _q_0 = uz_int_2  # q_0=20

    uz_new = uz_int_2 - (_q_0)

    lz_int_1 = lz_old + np.min(
        [perc, uz_int_1]
    )  # if the percolation > upper zone Q all the Quz will percolate

    _q_1 = k1 * lz_int_1

    if _q_1 > lz_int_1:
        _q_1 = lz_int_1

    lz_new = lz_int_1 - (_q_1)

    q_new = ((_q_0 + _q_1) * area) / (
        3.6 * tfac
    )  # q mm , area sq km  (1000**2)/1000/f/60/60 = 1/(3.6*f)
    # if daily tfac=24 if hourly tfac=1 if 15 min tfac=0.25
    return q_new, uz_new, lz_new


def Lake(temp, curve, tfac, rf, sf, q_new, lv_old, ltt, c_le, ep, lakeA):
    # lower zone
    # explicit representation of the lake where lake will be represented by a rating curve
    """
    l_ea :lake evaporation
    c_le : lake _evaporation correction factor

    """
    # lake evaporation
    if temp >= ltt:
        l_ea = (
            c_le * ep
        )  # the evaporation will be the potential evapotranspiration times correction factor
    else:
        l_ea = 0  # Evaporation will not occur when the Temperature is below the Threshold temperature

    l_ea_vol = l_ea * lakeA * 1000  # evaporation volume m3/time step

    # evaporation on the lake
    if temp < ltt:
        l_p = sf * c_le
    else:
        l_p = rf * c_le

    l_p_vol = l_p * lakeA * 1000  # prec # precipitation volume/ timestep

    q_vol = q_new * 3600 * tfac  # volume of inflow to the lake

    # storage in the lake before calculating the outflow
    lkv1 = lv_old + l_p_vol + q_vol - l_ea_vol
    # average storage for interpolation
    lkv2 = (lkv1 + lv_old) / 2

    storage = curve[:, 1]
    discharge = curve[:, 0]
    fn = interp11(storage, discharge, k=1)
    qout = max(fn(lkv2).tolist(), 0)

    lv_new = lkv2 - (qout * 3600 * tfac)

    return qout, lv_new


def Step_run(p, p2, v, St, curve, lake_sim, snow=0):
    """Step run.

    Makes the calculation of next step of discharge and states

    Parameters
    ----------
    p : array_like [10]
        Parameter vector, set up as:
        [fc, beta, e_corr, etf, lp, c_flux, k, k1, alpha, perc]
        9 fixed parameters + 9 optimized parameters + 1 parameter for the lake
    p2 : array_like [3]
        Problem parameter vector setup as:
        [tfac, area, Lake_area]
    v : array_like [4]
        Input vector setup as:
        [prec, temp, evap, llt]
    St : array_like [5]
        Previous model states setup as:
        [sp, sm, uz, lz, wc]
    Returns
    -------
    q_new : float
        Discharge [m3/s]
    St : array_like [5]
        Posterior model states
    """
    ## Parse of parameters from input vector to model
    # picipitation function
    if snow == 1:
        assert len(p) == 18, (
            "current version of HBV (with snow) takes 18 parameter you have entered "
            + str(len(p))
        )
        ltt = p[0]
        utt = p[1]
        rfcf = p[2]
        sfcf = p[3]
        # snow function
        ttm = p[4]
        cfmax = p[5]
        cwh = p[6]
        cfr = p[7]
        # soil function
        fc = p[8]
        beta = p[9]
        e_corr = [10]
        etf = p[11]
        lp = p[12]
        c_flux = p[13]
        # response function
        k = p[14]
        k1 = p[15]
        alpha = p[16]
        perc = p[17]
    #        pcorr=p[18]

    elif snow == 0:
        ltt = 1.0  # less than utt and less than lowest temp to prevent sf formation
        utt = (
            2.0  # very low but it does not matter as temp is 25 so it is greater than 2
        )
        rfcf = p[0]  # 1.0 #p[16] # all precipitation becomes rainfall
        sfcf = 0.00001  # there is no snow
        # snow function
        ttm = 1  # should be very low lower than lowest temp as temp is 25 all the time so it does not matter
        cfmax = 0.00001  # as there is no melting  and sp+sf=zero all the time so it doesn't matter the value of cfmax
        cwh = 0.00001  # as sp is always zero it doesn't matter all wc will go as inf
        cfr = 0.000001  # as temp > ttm all the time so it doesn't matter the value of cfr but put it zero
        # soil function
        fc = p[1]
        beta = p[2]
        e_corr = 1  # p[2]
        etf = p[3]
        lp = p[4]
        c_flux = p[5]
        # response function
        k = p[6]
        k1 = p[7]
        alpha = p[8]
        perc = p[9]

    ## Non optimisable parameters
    # [tfac,main_catchment_area,lake_sub,lake_area]
    tfac = p2[0]

    ## Parse of Inputs
    prec = v[0]  # Precipitation [mm]
    temp = v[1]  # Temperature [C]
    ep = v[2]
    tm = v[3]  # Long term (monthly) average temperature [C]

    ## Parse of states
    sp_old = St[0]
    sm_old = St[1]
    uz_old = St[2]
    lz_old = St[3]
    wc_old = St[4]

    if lake_sim:
        area = p2[2]  # area of lake subcatchment
        lakeA = p2[3]  # area of the lake
        c_le = p[9]
        lv_old = St[5]
    else:
        area = p2[1]

    rf, sf = Precipitation(temp, ltt, utt, prec, rfcf, sfcf, tfac)
    inf, wc_new, sp_new = Snow(cfmax, tfac, temp, ttm, cfr, cwh, rf, sf, wc_old, sp_old)
    sm_new, uz_int_1 = Soil(
        fc, beta, etf, temp, tm, e_corr, lp, tfac, c_flux, inf, ep, sm_old, uz_old
    )
    q_new, uz_new, lz_new = Response(tfac, perc, alpha, k, k1, area, lz_old, uz_int_1)

    # if lake_sim is true it will enter the function of the lake
    if lake_sim:
        qout, lv_new = Lake(
            temp, curve, tfac, rf, sf, q_new, lv_old, ltt, c_le, ep, lakeA
        )
    else:  # if lake_sim is false it will enter the function of the lake
        qout = q_new
        lv_new = 0

    return qout, [sp_new, sm_new, uz_new, lz_new, wc_new, lv_new]


def Simulate(
    prec,
    temp,
    et,
    par,
    p2,
    init_st=None,
    ll_temp=None,
    q_init=None,
    snow=0,
    lake_sim=False,
    curve=None,
):
    """Simulate.

    Run the HBV model for the number of steps (n) in precipitation. The
    resluts are (n+1) simulation of discharge as the model calculates step n+1

    Parameters
    ----------
    avg_prec : array_like [n]
        Average precipitation [mm/h]
    temp : array_like [n]
        Average temperature [C]
    et : array_like [n]
        Potential Evapotranspiration [mm/h]
    par : array_like [10]
        Parameter vector, set up as:
        [fc, beta, e_corr, etf, lp, c_flux, k, k1, alpha, perc]
    p2 : array_like [3]
        Problem parameter vector setup as:
        [tfac, area, Lake_area]
    init_st : array_like [5], optional
        Initial model states, [sp, sm, uz, lz, wc]. If unspecified,
        [0.0, 30.0, 30.0, 30.0, 0.0] mm
    ll_temp : array_like [n], optional
        Long term average temptearature. If unspecified, calculated from temp.
    q_0 : float, optional
        Initial discharge value. If unspecified set to 10.0

    Returns
    -------
    q_sim : array_like [n]
        Discharge for the n time steps of the precipitation vector [m3/s]
    st : array_like [n, 5]
        Model states for the complete time series [mm]
    """
    ### inputs validation
    # data type
    assert (
        len(init_st) == 5
    ), "state variables are 5 and the given initial values are " + str(len(init_st))
    assert type(p2) == list, " p2 should be of type list"
    assert len(p2) == 2, "p2 should contains tfac and catchment area"
    assert (
        snow == 0 or snow == 1
    ), " snow input defines whether to consider snow subroutine or not it has to be 0 or 1"

    if init_st is None:  # If unspecified, [0.0, 30.0, 30.0, 30.0, 0.0] mm
        st = [DEF_ST]  # if not given take the default
    else:
        st = [init_st]  # if given take it

    if (
        ll_temp is None
    ):  # If Long term average temptearature unspecified, calculated from temp
        ll_temp = [np.mean(temp)] * len(prec)

    if q_init is None:
        q_sim = [DEF_q0]
    else:
        q_sim = [q_init]

    # run the step by step function
    for i in range(len(prec)):
        v = [prec[i], temp[i], et[i], ll_temp[i]]
        q_out, st_out = Step_run(par, p2, v, st[i], curve, lake_sim, snow=0)
        q_sim.append(q_out)
        st.append(st_out)

    return q_sim, st



"""HBV model with a lake function."""
from __future__ import division, print_function

import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline as interp11

# initial values for state variables
# [sp, sm, uz, lz, wc]
DEF_ST = [0.0, 10.0, 10.0, 10.0, 0.0, 10.163 * 10**9]
# initial value for discarge
DEF_q0 = 2.3  # 10.0


def _precipitation(temp, ltt, utt, prec, rfcf, sfcf, tfac, pcorr):
    """Precipitation.
        inputs:
            1-precipitation
            2- Temperature
        outputs:
            1- rainfall
            2- snowfall
        Precipitaiton routine of the HBV96 model.
        If temperature is lower than ltt, all the precipitation is considered as
        snow. If the temperature is higher than utt, all the precipitation is
        considered as rainfall. In case that the temperature is between ltt and
        utt, precipitation is a linear mix of rainfall and snowfall.

    Parameters
    ----------
    temp : float
        Measured temperature [C]
    ltt : float
        Lower temperature treshold [C]
    utt : float
        Upper temperature treshold [C]
    prec : float
        Precipitation [mm]
    rfcf : float
        Rainfall corrector factor
    sfcf : float
        Snowfall corrector factor
    Returns
    -------
    _rf : float
        Rainfall [mm]
    _sf : float
        Snowfall [mm]
    """

    if temp <= ltt:  # if temp <= lower temp threshold
        _rf = 0.0  # no rainfall all the precipitation will convert into snowfall
        _sf = prec * sfcf

    elif temp >= utt:  # if temp > upper threshold
        _rf = prec * rfcf  # no snowfall all the precipitation becomes rainfall
        _sf = 0.0

    else:  # if  ltt< temp < utt
        _rf = ((temp - ltt) / (utt - ltt)) * prec * rfcf
        _sf = (1.0 - ((temp - ltt) / (utt - ltt))) * prec * sfcf

    _rf = _rf * pcorr

    return _rf, _sf


def _snow(cfmax, tfac, temp, ttm, cfr, cwh, _rf, _sf, wc_old, sp_old):
    """Snow. Snow routine of the HBV-96 model.

        The snow pack consists of two states: Water Content (wc) and Snow Pack
        (sp). The first corresponds to the liquid part of the water in the snow,
        while the latter corresponds to the solid part. If the temperature is
        higher than the melting point, the snow pack will melt and the solid snow
        will become liquid. In the opposite case, the liquid part of the snow will
        refreeze, and turn into solid. The water that cannot be stored by the solid
        part of the snow pack will drain into the soil as part of infiltration.

    Parameters
    ----------
    cfmax : float
        Day degree factor   # melting factor
    tfac : float
        Temperature correction factor
    temp : float
        Temperature [C]
    ttm : float
        Temperature treshold for Melting [C]
    cfr : float
        Refreezing factor
    cwh : float
        Capacity for water holding in snow pack
        water-holding capacity of snow (meltwater is retained in snowpack until it exceeds the WHC)
    _rf : float
        Rainfall [mm]
    _sf : float
        Snowfall [mm]
    wc_old : float
        Water content in previous state [mm]
    sp_old : float
        snow pack in previous state [mm]
    Returns
    -------
    _in : float
        Infiltration [mm]
    _wc_new : float
        Water content in posterior state [mm]
    _sp_new : float
        Snowpack in posterior state [mm]
    """

    if temp > ttm:  # if temp > melting threshold
        # then either some snow will melt or the entire snow will melt
        if (
            cfmax * (temp - ttm) < sp_old + _sf
        ):  # if amount of melted snow < the entire existing snow (previous amount+new)
            _melt = cfmax * (temp - ttm)
        else:  # if amount of melted snow > the entire existing snow (previous amount+new)
            _melt = (
                sp_old + _sf
            )  # then the entire existing snow will melt (old snow pack + the current snowfall)

        _sp_new = sp_old + _sf - _melt
        _wc_int = wc_old + _melt + _rf

    else:  # if temp < melting threshold
        # then either some water will freeze or all the water willfreeze
        if (
            cfr * cfmax * (ttm - temp) < wc_old + _rf
        ):  # if the amount of frozen water < entire water available
            _refr = (
                cfr * cfmax * (ttm - temp)
            )  # cfmax*(ttm-temp) is the rate of melting of snow while cfr*cfmax*(ttm-temp) is the rate of freeze of melted water  (rate of freezing > rate of melting)
        else:  # if the amount of frozen water > entire water available
            _refr = wc_old + _rf

        _sp_new = sp_old + _sf + _refr
        _wc_int = wc_old - _refr + _rf

    if _wc_int > cwh * _sp_new:  # if water content > holding water capacity of the snow
        _in = _wc_int - cwh * _sp_new  # water content  will infiltrate
        _wc_new = (
            cwh * _sp_new
        )  # and the capacity of snow of holding water will retained
    else:  # if water content < holding water capacity of the snow
        _in = 0.0  # no infiltration
        _wc_new = _wc_int

    return _in, _wc_new, _sp_new


def _soil(fc, beta, etf, temp, tm, e_corr, lp, tfac, c_flux, inf, ep, sm_old, uz_old):
    """Soil. Soil routine of the HBV-96 model.

        The model checks for the amount of water that can infiltrate the soil,
        coming from the liquid precipitation and the snow pack melting. A part of
        the water will be stored as soil moisture, while other will become runoff,
        and routed to the upper zone tank.

    Parameters
    ----------
    fc : float
        Filed capacity
    beta : float
        Shape coefficient for effective precipitation separation
        Controls the contribution of the increase in the soil moisture or
        to the response function
    etf : float
        Total potential evapotranspiration
    temp : float
        Temperature
    tm : float
        Average long term temperature
    e_corr : float
        Evapotranspiration corrector factor
    lp : float _soil

    tfac : float
        Time conversion factor
    c_flux : float
        Capilar flux in the root zone
    _in : float
        actual infiltration
    ep : float
        Long term mean potential evapotranspiration
    sm_old : float
        Previous soil moisture value
    uz_old : float
        Previous Upper zone value
    Returns
    -------
    sm_new : float
        New value of soil moisture
    uz_int_1 : float
        New value of direct runoff into upper zone

    variables
    -------
    EPInt  :
        Adjusted potential evapotranspiration
    """

    #    qdr = max(sm_old + inf - fc, 0)  # direct run off as soil moisture exceeded the field capacity
    qdr = 0
    _in = inf - qdr
    _r = ((sm_old / fc) ** beta) * _in  # recharge from soil subroutine to upper zone

    _ep_int = max(
        (1.0 + etf * (temp - tm)) * e_corr * ep, 0
    )  # Adjusted potential evapotranspiration

    _ea = min(_ep_int, (sm_old / (lp * fc)) * _ep_int)

    _cf = c_flux * ((fc - sm_old) / fc)  # capilary rise

    # if capilary rise is more than what is available take all the available and leave it empty

    if uz_old + _r < _cf:
        _cf = uz_old + _r
        uz_int_1 = 0
    else:
        uz_int_1 = uz_old + _r - _cf

    #    uz_int_1 = max(uz_old + _r - _cf,0)

    sm_new = max(sm_old + _in - _r + _cf - _ea, 0)

    return sm_new, uz_int_1, qdr


def _response(tfac, perc, alpha, k, k1, area, lz_old, uz_int_1, qdr):
    """Response. The response routine of the HBV-96 model.

        The response routine is in charge of transforming the current values of
        upper and lower zone into discharge. This routine also controls the
        recharge of the lower zone tank (baseflow). The transformation of units
        also occurs in this point.

    Parameters
    ----------
    tfac : float
        Number of hours in the time step
    perc : float
        Percolation value [mm\hr]
    alpha : float
        Response box parameter
        upper zone runoff coefficient
    k : float
        Upper zone recession coefficient
        Upper zone response coefficient
    k1 : float
        Lower zone recession coefficient
        Lowe zone response coefficient
    area : float
        Catchment area [Km2]
    lz_old : float
        Previous lower zone value [mm]
    uz_int_1 : float
        Previous upper zone value before percolation [mm]
    qdr : float
        Direct runoff [mm]
    """
    # upper zone
    # if perc > Quz then perc = Quz and Quz = 0 if not perc = value and Quz= Quz-perc so take the min
    uz_int_2 = np.max([uz_int_1 - perc, 0.0])  # upper zone after percolation
    _q_0 = k * (uz_int_2 ** (1.0 + alpha))

    if _q_0 > uz_int_2:  # if q_0 =30 and UZ=20
        _q_0 = uz_int_2  # q_0=20

    uz_new = uz_int_2 - _q_0

    lz_int_1 = lz_old + np.min(
        [perc, uz_int_1]
    )  # if the percolation > upper zone Q all the Quz will percolate

    _q_1 = k1 * lz_int_1

    if _q_1 > lz_int_1:
        _q_1 = lz_int_1

    lz_new = lz_int_1 - _q_1

    q_new = ((_q_0 + _q_1) * area) / (
        3.6 * tfac
    )  # q mm , area sq km  (1000**2)/1000/f/60/60 = 1/(3.6*f)
    # if daily tfac=24 if hourly tfac=1 if 15 min tfac=0.25
    return q_new, uz_new, lz_new


def _lake(temp, curve, tfac, rf, sf, q_new, lv_old, ltt, c_le, ep, lakeA):
    """lake. lake model.

    Parameters
    ----------
    temp
    curve
    tfac
    rf
    sf
    q_new
    lv_old
    ltt
    c_le
    ep
    lakeA

    Returns
    -------
    """
    # lower zone
    # explicit representation of the lake where lake will be represented by a rating curve
    """
    l_ea :lake evaporation
    c_le : lake _evaporation correction factor

    """
    # lake evaporation
    if temp >= ltt:
        l_ea = (
            c_le * ep
        )  # the evaporation will be the potential evapotranspiration times correction factor
    else:
        l_ea = 0  # Evaporation will not occur when the Temperature is below the Threshold temperature

    l_ea_vol = l_ea * lakeA * 1000  # evaporation volume m3/time step

    # evaporation on the lake
    if temp < ltt:
        l_p = sf * c_le
    else:
        l_p = rf * c_le

    l_p_vol = l_p * lakeA * 1000  # prec # precipitation volume/ timestep

    q_vol = q_new * 3600 * tfac  # volume of inflow to the lake

    # storage in the lake before calculating the outflow
    lkv1 = lv_old + l_p_vol + q_vol - l_ea_vol
    # average storage for interpolation
    lkv2 = (lkv1 + lv_old) / 2

    storage = curve[:, 1]
    discharge = curve[:, 0]
    fn = interp11(storage, discharge, k=1)
    qout = max(fn(lkv2).tolist(), 0)

    lv_new = lkv2 - (qout * 3600 * tfac)

    return qout, lv_new


def _tf(maxbas):
    """tf.

    Transfer function weight generator in a shape of a triangle.
    """

    wi = []
    for x in range(1, maxbas + 1):  # if maxbas=3 so x=[1,2,3]
        if (
            x <= maxbas / 2.0
        ):  # x <= 1.5  # half of values will form the rising limb and half falling limb
            # Growing transfer    # rising limb
            wi.append(x / (maxbas + 2.0))
        else:
            # Receding transfer    # falling limb
            wi.append(1.0 - (x + 1) / (maxbas + 2.0))

    # Normalise weights
    wi = np.array(wi) / np.sum(wi)
    return wi


def _routing(q, maxbas=1):
    """routing.

    This function implements the transfer function using a triangular
    function.
    """
    assert maxbas >= 1, "Maxbas value has to be larger than 1"
    # Get integer part of maxbas
    maxbas = int(round(maxbas, 0))

    # get the weights
    w = _tf(maxbas)

    # rout the discharge signal
    q_r = np.zeros_like(q, dtype="float64")
    q_temp = q
    for w_i in w:
        q_r += q_temp * w_i
        q_temp = np.insert(q_temp, 0, 0.0)[:-1]

    return q_r


def CalculateMaxBas(MAXBAS):
    """CalculateMaxBas. This function calculate the MAXBAS Weights based on a MAXBAX number The MAXBAS is a HBV parameter that controls the routing.

    Parameters
    ----------
    MAXBAS: [int]
        maxbas weight

    Example
    -------
    maxbasW = CalculateMaxBas(5)
    maxbasW =

        0.0800    0.2400    0.3600    0.2400    0.0800

    It is important to mention that this function allows to obtain weights
    not only for interger values but from decimals values as well.
    """
    yant = 0
    Total = 0  # Just to verify how far from the unit is the result

    TotalA = (MAXBAS * MAXBAS * np.sin(np.pi / 3)) / 2

    IntPart = np.floor(MAXBAS)

    RealPart = MAXBAS - IntPart

    PeakPoint = MAXBAS % 2

    flag = 1  # 1 = "up"  ; 2 = down

    if RealPart > 0:  # even number 2,4,6,8,10
        maxbasW = np.ones(int(IntPart) + 1)  # if even add 1
    else:  # odd number
        maxbasW = np.ones(int(IntPart))

    for x in range(int(MAXBAS)):

        if x < (MAXBAS / 2.0) - 1:
            ynow = np.tan(np.pi / 3) * (x + 1)
            # Integral of  x dx with slope of 60 degree Equilateral triangle
            maxbasW[x] = ((ynow + yant) / 2) / TotalA  # ' Area / Total Area

        else:  # The area here is calculated by the formlua of a trapezoidal (B1+B2)*h /2
            if flag == 1:
                ynow = np.sin(np.pi / 3) * MAXBAS
                if PeakPoint == 0:
                    maxbasW[x] = ((ynow + yant) / 2) / TotalA
                else:
                    A1 = ((ynow + yant) / 2) * (MAXBAS / 2.0 - x) / TotalA
                    yant = ynow
                    ynow = (MAXBAS * np.sin(np.pi / 3)) - (
                        np.tan(np.pi / 3) * (x + 1 - MAXBAS / 2.0)
                    )
                    A2 = ((ynow + yant) * (x + 1 - MAXBAS / 2.0) / 2) / TotalA
                    maxbasW[x] = A1 + A2

                flag = 2
            else:
                ynow = MAXBAS * np.sin(np.pi / 3) - np.tan(np.pi / 3) * (
                    x + 1 - MAXBAS / 2.0
                )  # 'sum of the two height in the descending part of the triangle
                maxbasW[x] = ((ynow + yant) / 2) / TotalA
                # Multiplying by the height of the trapezoidal and dividing by 2

        Total = Total + maxbasW[x]
        yant = ynow

    x = x + 1

    if RealPart > 0:
        if np.floor(MAXBAS) == 0:
            MAXBAS = 1
            maxbasW[x] = 1
            NumberofWeights = 1
        else:
            maxbasW[x] = (yant * (MAXBAS - x) / 2) / TotalA
            Total = Total + maxbasW[x]
            NumberofWeights = x
    else:

        NumberofWeights = x - 1

    return maxbasW


def RoutingMAXBAS(Q, MAXBAS):
    """RoutingMAXBAS. This function calculate the routing from a input hydrograph using the MAXBAS parameter from the HBV model.

    Parameters
    ---------
    Q: [array]
        input hydrograph
    MAXBAS: []
        MAXBAS weight
    Qout = output hydrograph
    """
    # CALCULATE MAXBAS WEIGHTS
    maxbasW = CalculateMaxBas(MAXBAS)
    Qw = np.ones((len(Q), len(maxbasW)))
    # Calculate the matrix discharge
    for i in range(len(Q)):  # 0 to 10
        for k in range(len(maxbasW)):  # 0 to 4
            Qw[i, k] = maxbasW[k] * Q[i]

    def mm(A, s):
        tot = []
        for o in range(np.shape(A)[1]):  # columns
            for t in range(np.shape(A)[0]):  # rows
                tot.append(A[t, o])
        Su = tot[s:-1:s]
        return Su

    # Calculate routing
    j = 0
    Qout = np.ones((len(Q), 1))

    for i in range(len(Q)):
        if i == 0:
            Qout[i] = Qw[i, i]
        elif i < len(maxbasW) - 1:
            A = Qw[0 : i + 1, :]
            s = len(A) - 1  # len(A) is the no of rows or use int(np.shape(A)[0])
            Su = mm(A, s)

            Qout[i] = sum(Su[0 : i + 1])
        else:
            A = Qw[j : i + 1, :]
            s = len(A) - 1
            Su = mm(A, s)
            Qout[i] = sum(Su)
            j = j + 1

    #        del A ,s, Su

    return Qout  # ,maxbasW


def _step_run(p, p2, v, St, curve):
    """Step run. Makes the calculation of next step of discharge and states.

    Parameters
    ----------
    p : array_like [10]
        Parameter vector, set up as:
        [fc, beta, e_corr, etf, lp, c_flux, k, k1, alpha, perc]
        9 fixed parameters + 9 optimized parameters + 1 parameter for the lake
    p2 : array_like [3]
        Problem parameter vector setup as:
        [tfac, area, Lake_area]
    v : array_like [4]
        Input vector setup as:
        [prec, temp, evap, llt]
    St : array_like [5]
        Previous model states setup as:
        [sp, sm, uz, lz, wc]

    Returns
    -------
    q_new : float
        Discharge [m3/s]
    St : array_like [5]
        Posterior model states
    """
    ## Parse of parameters from input vector to model
    # picipitation function
    ltt = 1.0  # p[0] # less than utt and less than lowest temp to prevent sf formation
    utt = 2.0  # p[1]  #very low but it does not matter as temp is 25 so it is greater than 2
    rfcf = 1.0  # p[16] # all precipitation becomes rainfall
    sfcf = 0.00001  # p[17] # there is no snow
    # snow function
    ttm = 1  # p[2] #should be very low lower than lowest temp as temp is 25 all the time so it does not matter
    cfmax = 0.00001  # p[3] as there is no melting  and sp+sf=zero all the time so it doesn't matter the value of cfmax
    cwh = 0.00001  # p[12] as sp is always zero it doesn't matter all wc will go as inf
    cfr = 0.000001  # p[13] as temp > ttm all the time so it doesn't matter the value of cfr but put it zero
    # soil function
    fc = p[0]
    beta = p[1]
    e_corr = 1  # p[2]
    etf = p[2]
    lp = p[3]
    c_flux = p[4]
    # response function
    k = p[5]
    k1 = p[6]
    alpha = p[7]
    perc = p[8]

    ## Non optimisable parameters
    # [tfac,jiboa,lake_sub,lake_area]
    tfac = p2[0]  # tfac=0.25
    # jiboa_area=p2[1] # AREA = 432
    lake_sub = p2[1]  # area of lake subcatchment
    lakeA = p2[2]  # area of the lake
    #    ilake=lakeA/area  # percentage of the lake area to catchment area

    ## Parse of Inputs
    avg_prec = v[0]  # Precipitation [mm]
    temp = v[1]  # Temperature [C]
    ep = v[2]
    tm = v[3]  # Long term (monthly) average temperature [C]

    ## Parse of states
    sp_old = St[0]
    sm_old = St[1]
    uz_old = St[2]
    lz_old = St[3]
    wc_old = St[4]

    # if lake_sim:
    area = lake_sub
    c_le = p[9]
    lv_old = St[5]
    pcorr = p[10]
    # else:
    # area=jiboa_area
    # pcorr=p[9]

    #    pcorr=p[10]

    rf, sf = _precipitation(temp, ltt, utt, avg_prec, rfcf, sfcf, tfac, pcorr)
    inf, wc_new, sp_new = _snow(
        cfmax, tfac, temp, ttm, cfr, cwh, rf, sf, wc_old, sp_old
    )
    sm_new, uz_int_1, qdr = _soil(
        fc, beta, etf, temp, tm, e_corr, lp, tfac, c_flux, inf, ep, sm_old, uz_old
    )
    q_new, uz_new, lz_new = _response(
        tfac, perc, alpha, k, k1, area, lz_old, uz_int_1, qdr
    )

    # if lake_sim is true it will enter the function of the lake
    # if lake_sim:
    qout, lv_new = _lake(temp, curve, tfac, rf, sf, q_new, lv_old, ltt, c_le, ep, lakeA)
    # else:    # if lake_sim is false it will enter the function of the lake
    # qout=q_new
    # lv_new=0

    return qout, [sp_new, sm_new, uz_new, lz_new, wc_new, lv_new]


def simulate(
    avg_prec,
    temp,
    et,
    par,
    p2,
    curve,
    q_0=DEF_q0,
    init_st=None,
    ll_temp=None,
    lake_sim=False,
):
    """Simulate.

        Run the HBV model for the number of steps (n) in precipitation. The
        resluts are (n+1) simulation of discharge as the model calculates step n+1

    Parameters
    ----------
    avg_prec : array_like [n]
        Average precipitation [mm/h]
    temp : array_like [n]
        Average temperature [C]
    et : array_like [n]
        Potential Evapotranspiration [mm/h]
    par : array_like [10]
        Parameter vector, set up as:
        [fc, beta, e_corr, etf, lp, c_flux, k, k1, alpha, perc]
    p2 : array_like [3]
        Problem parameter vector setup as:
        [tfac, area, Lake_area]
    init_st : array_like [5], optional
        Initial model states, [sp, sm, uz, lz, wc]. If unspecified,
        [0.0, 30.0, 30.0, 30.0, 0.0] mm
    ll_temp : array_like [n], optional
        Long term average temptearature. If unspecified, calculated from temp.
    q_0 : float, optional
        Initial discharge value. If unspecified set to 10.0

    Returns
    -------
    q_sim : array_like [n]
        Discharge for the n time steps of the precipitation vector [m3/s]
    st : array_like [n, 5]
        Model states for the complete time series [mm]
    """

    if init_st is None:  # If unspecified, [0.0, 30.0, 30.0, 30.0, 0.0] mm
        st = [DEF_ST]  # if not given take the default
    else:
        st = [init_st]  # if given take it

    if (
        ll_temp is None
    ):  # If Long term average temptearature unspecified, calculated from temp
        ll_temp = [np.mean(temp)] * len(avg_prec)

    # q_sim = [q_0, ]
    q_sim = []

    for i in range(len(avg_prec)):
        v = [avg_prec[i], temp[i], et[i], ll_temp[i]]
        q_out, st_out = _step_run(par, p2, v, st[i], curve)
        q_sim.append(q_out)
        st.append(st_out)

    return q_sim, st



"""Created on Wed Jul 18 23:55:20 2018.

@author: Mostafa
"""
import datetime
import pickle


def save_obj(obj, saved_name):
    """
    ===============================================================
        save_obj(obj, saved_name )
    ===============================================================
    this function is used to save any python object to your hard desk

    Inputs:
    ----------
        1-obj:

        2-saved_name:
            ['String'] name of the object
    Outputs:
    ----------
        the object will be saved to the given path/current working directory
        with the given name
    Example:
        data={"key1":[1,2,3,5],"key2":[6,2,9,7]}
        save_obj(data,path+'/flow_acc_table')
    """
    with open(saved_name + ".pkl", "wb") as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)


def load_obj(saved_name):
    """
    ===============================================================
        load_obj(saved_name)
    ===============================================================
    this function is used to save any python object to your hard desk

    Inputs:
    ----------
        1-saved_name:
            ['String'] name of the object
    Outputs:
    ----------
        the object will be loaded
    Example:
        load_obj(path+'/flow_acc_table')
    """
    with open(saved_name + ".pkl", "rb") as f:
        return pickle.load(f)


def dateformated(x):
    """dateformated.

    this function converts the the date read from a list to a datetime format

    Parameters
    ----------
    x: [list]
        is a list of tuples of string date read from database

    Returns
    -------
        list od dates as a datetime format  YYYY-MM-DD HH:MM:SS
    """
    x = [i[0] for i in x]
    #
    x1 = []
    for i in x:
        if len(i) == 19:
            x1.append(
                datetime.datetime(
                    int(i[:4]),
                    int(i[5:7]),
                    int(i[8:10]),
                    int(i[11:13]),
                    int(i[14:16]),
                    int(i[17:18]),
                )
            )
    #        elif len(i)==13:
    #            x1.append(datetime.datetime(int(i[:4]),int(i[5:7]),int(i[8:10]),int(i[11:13]),int(0),int(0) ))
    #        else:
    #            x1.append(datetime.datetime(int(i[:4]),int(i[5:7]),int(i[8:10]),int(0),int(0),int(0) ))
    #    del i,x
    return x1


def printWaitBar(i, total, prefix="", suffix="", decimals=1, length=100, fill="█"):
    """This function will print a waitbar in the console.

    Variables:

    i -- Iteration number
    total -- Total iterations
    fronttext -- Name in front of bar
    prefix -- Name after bar
    suffix -- Decimals of percentage
    length -- width of the waitbar
    fill -- bar fill
    """
    import os
    import sys

    # Adjust when it is a linux computer
    if os.name == "posix" and total == 0:
        total = 0.0001

    percent = ("{0:." + str(decimals) + "f}").format(100 * (i / float(total)))
    filled = int(length * i // total)
    bar = fill * filled + "-" * (length - filled)

    sys.stdout.write("\r%s |%s| %s%% %s" % (prefix, bar, percent, suffix))
    sys.stdout.flush()

    if i == total:
        print()



"""Created on Tue May 22 09:01:55 2018.

@author: Mostafa
"""
try:
    from importlib.metadata import PackageNotFoundError, version  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version  # type: ignore


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

# documentation format
__author__ = "Mostafa Farrag"
__email__ = "moah.farag@gmail.com"
__docformat__ = "restructuredtext"

# Let users know if they're missing any of our hard dependencies
hard_dependencies = ()  # ("numpy", "pandas", "gdal")
missing_dependencies = []

for dependency in hard_dependencies:
    try:
        __import__(dependency)
    except ImportError as e:
        missing_dependencies.append(dependency)
        print(e)

if missing_dependencies:
    raise ImportError("Missing required dependencies {0}".format(missing_dependencies))


# del hard_dependencies, dependency, missing_dependencies


def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration(None, parent_package, top_path)
    config.set_options(
        ignore_setup_xxx_py=True,
        assume_default_configuration=True,
        delegate_options_to_subpackages=True,
        quiet=True,
    )

    # config.add_subpackage("gis")
    config.add_subpackage("hm")
    config.add_subpackage("plot")
    config.add_subpackage("rrm")
    # config.add_subpackage("rs")
    # config.add_subpackage("sm")
    return config


# import Hapi.catchment
# import Hapi.hm
# import Hapi.java_functions
# import Hapi.plot
# import Hapi.rrm
# import Hapi.run
# import Hapi.weirdFn

# import Hapi.saintvenant
# module level doc-string
__doc__ = """
Hapi - Hydrological library for Python
=====================================================================

**Hapi** is a Python package providing fast and flexible, way to build distributed
hydrological model using lumped conceptual model

Main Features
-------------
Here are just a few of the things that pandas does well:

  - Easy handling of rasters data downloaded from global data and easy way to
    manipulate the data to arrange it to run the model
  - Easy calibration of the model using Harmony search method and Genetic Algorithms
  - flexible GIS function to process rasters interpolate values and georeference
   calculated discharge values to the correct place

"""
# if __name__=='__main__':
#     print("Hapi")



"""Created on Wed May 16 03:50:00 2018.

@author: Mostafa
"""
import datetime as dt
import os
import shutil

import numpy as np
import pandas as pd
import rasterio
from loguru import logger
from osgeo import gdal
from pyramids.raster import Raster as raster
from rasterio.plot import show

# from datetime import datetime
# import geopandas as gpd
from rasterstats import zonal_stats

import Hapi


class Inputs:
    """Inputs.

    Inputs class contains methods to prepare the inputs for the distributed
    hydrological model

    Methods:
        1- prepareInputs
        2- extractParametersBoundaries
        3- extractParameters
        4- createLumpedInputs
        5- renameFiles
        6- changetext2time
        7- ReadExcelData
        8- ListAttributes
    """

    def __init__(self):
        """no parameters needed for instantiating the object."""
        pass

    def prepareInputs(src: str, input_folder: str, folder_name: str):
        """prepareInputs.

        this function prepare downloaded raster data to have the same align and
        nodatavalue from a GIS raster (DEM, flow accumulation, flow direction raster)
        and return a folder with the output rasters with a name "New_Rasters"

        Parameters
        ----------
        src: [str]
            path to the spatial information source raster to get the spatial information
            (coordinate system, no of rows & columns) A_path should include the name of the raster
            and the extension like "data/dem.tif".
        input_folder: [str]
            path of the folder of the rasters you want to adjust their
            no of rows, columns and resolution (alignment) like raster A
            the folder should not have any other files except the rasters.
        folder_name: [str]
            name to create a folder to store resulted rasters.

        Example
        -------
        Ex1:
            >>> dem_path="01GIS/inputs/4000/acc4000.tif"
            >>> prec_in_path="02Precipitation/CHIRPS/Daily/"
            >>> Inputs.prepareInputs(dem_path,prec_in_path,"prec")
        Ex2:
            >>> dem_path="01GIS/inputs/4000/acc4000.tif"
            >>> outputpath="00inputs/meteodata/4000/"
            >>> evap_in_path="03Weather_Data/evap/"
            >>> Inputs.prepareInputs(dem_path,evap_in_path,outputpath+"evap")
        """
        if not isinstance(folder_name, str):
            print("folder_name input should be string type")
        # create a new folder for new created alligned rasters in temp
        # check if you can create the folder
        try:
            os.makedirs(os.path.join(os.environ["TEMP"], "AllignedRasters"))
        except WindowsError:
            # if not able to create the folder delete the folder with the same name and create one empty
            shutil.rmtree(os.path.join(os.environ["TEMP"] + "/AllignedRasters"))
            os.makedirs(os.path.join(os.environ["TEMP"], "AllignedRasters"))

        temp = os.environ["TEMP"] + "/AllignedRasters/"

        # match alignment
        print(
            "- First alligned files will be created in a folder 'AllignedRasters' in the Temp folder in you "
            "environment variable"
        )
        raster.matchDataAlignment(src, input_folder, temp)
        # create new folder in the current directory for alligned and nodatavalue matched cells
        try:
            os.makedirs(os.path.join(os.getcwd(), folder_name))
        except (WindowsError, FileExistsError):
            raise FileExistsError(
                f"please The function is trying to create a folder with a name {folder_name}  "
                f"to complete the process if there is a folder with the same name please rename "
                f"it to other name"
            )
        # match nodata value
        print(
            "- Second matching NoDataValue from the DEM raster too all raster will be created in the outputpath"
        )
        raster.cropAlignedFolder(temp, src, f"{folder_name}/")
        # delete the processing folder from temp
        shutil.rmtree(temp)

    @staticmethod
    def extractParametersBoundaries(Basin):
        """extractParametersBoundaries.

        extractParametersBoundaries

        Parameters
        ----------
        Basin : [Geodataframe]
            gepdataframe of catchment polygon, make sure that the geodataframe contains
            one row only, if not merge all the polygons in the shapefile first.

        Returns
        -------
        UB : [list]
            list of the upper bound of the parameters.
        LB : [list]
            list of the lower bound of the parameters.

        the parameters are
            ["tt", "sfcf","cfmax","cwh","cfr","fc","beta",
             "lp","k0","k1","k2","uzl","perc", "maxbas"]
        """
        ParametersPath = os.path.dirname(Hapi.__file__)
        ParametersPath = ParametersPath + "/parameters"
        ParamList = [
            "01_tt",
            "02_rfcf",
            "03_sfcf",
            "04_cfmax",
            "05_cwh",
            "06_cfr",
            "07_fc",
            "08_beta",
            "09_etf",
            "10_lp",
            "11_k0",
            "12_k1",
            "13_k2",
            "14_uzl",
            "15_perc",
            "16_maxbas",
            "17_K_muskingum",
            "18_x_muskingum",
        ]

        raster_obj = rasterio.open(ParametersPath + "/max/" + ParamList[0] + ".tif")
        Basin = Basin.to_crs(crs=raster_obj.crs)
        # max values
        UB = list()
        for i in range(len(ParamList)):
            raster_obj = rasterio.open(ParametersPath + "/max/" + ParamList[i] + ".tif")
            array = raster_obj.read(1)
            affine = raster_obj.transform
            UB.append(
                zonal_stats(Basin, array, affine=affine, stats=["max"])[0]["max"]
            )  # stats=['min', 'max', 'mean', 'median', 'majority']

        # min values
        LB = list()
        for i in range(len(ParamList)):
            raster_obj = rasterio.open(ParametersPath + "/min/" + ParamList[i] + ".tif")
            array = raster_obj.read(1)
            affine = raster_obj.transform
            LB.append(zonal_stats(Basin, array, affine=affine, stats=["min"])[0]["min"])

        Par = pd.DataFrame(index=ParamList)

        Par["UB"] = UB
        Par["LB"] = LB
        # plot the given basin with the parameters raster
        ax = show((raster_obj, 1), with_bounds=True)
        Basin.plot(facecolor="None", edgecolor="blue", linewidth=2, ax=ax)
        # ax.set_xbound([Basin.bounds.loc[0,'minx']-10,Basin.bounds.loc[0,'maxx']+10])
        # ax.set_ybound([Basin.bounds.loc[0,'miny']-1, Basin.bounds.loc[0,'maxy']+1])

        return Par

    def extractParameters(self, gdf, scenario, as_raster=False, save_to=""):
        """extractParameters.

        extractParameters method extracts the parameter rasters at the location
        of the source raster, there are 12 set of parameters 10 sets of parameters
        (Beck et al., (2016)) and the max, min and average of all sets


        Beck, H. E., Dijk, A. I. J. M. van, Ad de Roo, Diego G. Miralles,
        T. R. M. & Jaap Schellekens, and L. A. B. (2016) Global-scale
        regionalization of hydrologic model parameters-Supporting materials
        3599–3622. doi:10.1002/2015WR018247.Received

        Parameters
        ----------
        gdf: [Geodataframe]
            gepdataframe of catchment polygon, make sure that the geodataframe contains
            one row only, if not merge all the polygons in the shapefile first.
        scenario: [str]
            name of the parameter set, there are 12 sets of parameters
            ["1","2","3","4","5","6","7","8","9","10","avg","max","min"]
        as_raster: [bool]
            Default is False.
        save_to: [str]
            path to a directory where you want to save the rasters.

        Returns
        -------
        Parameters : [list]
            list of the upper bound of the parameters.


        the parameters are
            ["tt", rfcf,"sfcf","cfmax","cwh","cfr","fc","beta",'etf'
             "lp","k0","k1","k2","uzl","perc", "maxbas",'K_muskingum',
             'x_muskingum']
        """
        ParametersPath = os.path.dirname(Hapi.__file__)
        ParametersPath = ParametersPath + "/parameters/" + scenario
        ParamList = [
            "01_tt",
            "02_rfcf",
            "03_sfcf",
            "04_cfmax",
            "05_cwh",
            "06_cfr",
            "07_fc",
            "08_beta",
            "09_etf",
            "10_lp",
            "11_k0",
            "12_k1",
            "13_k2",
            "14_uzl",
            "15_perc",
            "16_maxbas",
            "17_K_muskingum",
            "18_x_muskingum",
        ]

        if not as_raster:
            raster_obj = rasterio.open(f"{ParametersPath}/{ParamList[0]}.tif")
            gdf = gdf.to_crs(crs=raster_obj.crs)
            # max values
            Par = list()
            for i in range(len(ParamList)):
                raster_obj = rasterio.open(f"{ParametersPath}/{ParamList[i]}.tif")
                array = raster_obj.read(1)
                affine = raster_obj.transform
                Par.append(
                    zonal_stats(gdf, array, affine=affine, stats=["max"])[0]["max"]
                )  # stats=['min', 'max', 'mean', 'median', 'majority']

            # plot the given basin with the parameters raster

            # Plot DEM
            ax = show((raster_obj, 1), with_bounds=True)
            gdf.plot(facecolor="None", edgecolor="blue", linewidth=2, ax=ax)
            # ax.set_xbound([Basin.bounds.loc[0,'minx']-10,Basin.bounds.loc[0,'maxx']+10])
            # ax.set_ybound([Basin.bounds.loc[0,'miny']-1, Basin.bounds.loc[0,'maxy']+1])

            return Par
        else:
            Inputs.prepareInputs(gdf, f"{ParametersPath}/", save_to)

    @staticmethod
    def createLumpedInputs(Path):
        """createLumpedInputs.

        createLumpedInputs method generate a lumped parameters from
        distributed parameters by taking the average
        Parameters
        ----------
        Path : [str]
            path to folder that contains the parameter rasters.

        Returns
        -------
        data : [array]
            array contains the average values of the distributed parameters.
        """
        # data type
        assert type(Path) == str, "PrecPath input should be string type"
        # check wether the path exists or not
        assert os.path.exists(Path), Path + " you have provided does not exist"
        # check wether the folder has the rasters or not
        assert len(os.listdir(Path)) > 0, Path + " folder you have provided is empty"
        # read data
        data = raster.readRastersFolder(Path)
        # get the No data value from the first raster in the folder
        scr = gdal.Open(f"{Path}/{os.listdir(Path)[0]}")
        _, NoDataValue = raster.getRasterData(scr)
        data[data == NoDataValue] = np.nan

        data = np.nanmean(data, axis=0)
        data = data.mean(0)

        return data

    @staticmethod
    def renameFiles(
        path: str, prefix: str = "", fmt: str = "%Y.%m.%d", freq: str = "daily"
    ):
        """renameFiles.

        renameFiles method takes the path to a folder where you want to put a number
        at the begining of the raster names indicating the order of the raster based on
        its date

        Parameters
        ----------
        path : [str]
            path where the rasters are stored.
        prefix: [str]
            any string you want to add to the raster names, (i.e the dataset name precipitation_ecmwf). Default is "".
        fmt : [String], optional
            the format of the date. The default is '%Y.%m.%d'.
        freq: [str]
            Default is "daily".

        Returns
        -------
        files in the Path are going to have a new name including the order at
        the begining of the name.
        """
        if not os.path.exists(path):
            raise FileNotFoundError("The directory you have entered does not exist")

        files = os.listdir(path)
        # get only the tif files
        files = [i for i in files if i.endswith(".tif")]

        # get the date
        dates_str = [files[i].split("_")[-1][:-4] for i in range(len(files))]
        dates = [dt.datetime.strptime(dates_str[i], fmt) for i in range(len(files))]

        if freq == "daily":
            new_date_str = [
                str(i.year) + "_" + str(i.month) + "_" + str(i.day) for i in dates
            ]
        elif freq == "hourly":
            new_date_str = [
                str(i.year) + "_" + str(i.month) + "_" + str(i.day) + "_" + str(i.hour)
                for i in dates
            ]
        else:
            new_date_str = [
                f"{i.year}-{i.month}-{i.day}-{i.hour}-{i.minute}" for i in dates
            ]

        df = pd.DataFrame()
        df["files"] = files
        df["DateStr"] = new_date_str
        df["dates"] = dates
        df.sort_values("dates", inplace=True)
        df.reset_index(inplace=True)
        df["order"] = [i for i in range(len(files))]

        df["new_names"] = [
            f"{df.loc[i, 'order']}_{prefix}_{df.loc[i, 'DateStr']}.tif"
            for i in range(len(files))
        ]
        # rename the files
        for i in range(len(files)):
            os.rename(
                f"{path}/{df.loc[i, 'files']}", f"{path}/{df.loc[i, 'new_names']}"
            )

    @staticmethod
    def changetext2time(string):
        """changetext2time.

        this functions changes the date from a string to a date time
        format
        """
        time = dt.datetime(
            int(string[:4]),
            int(string[5:7]),
            int(string[8:10]),
            int(string[11:13]),
            int(string[14:16]),
            int(string[17:]),
        )
        return time

    @staticmethod
    def ReadExcelData(path, years, months):
        """ReadExcelData.

            this function reads data listed in excel sheet with years and months are
            listed as columns and days are listed in the first row
            year month 1 2 3 4 5 6 7 8 9 .....................31
            2012  1    5 6 2 6 8 6 9 7 4 3 ...................31
            2012  2    9 8 7 6 3 2 1 5 5 9 ...................31

        Parameters
        ----------
        path:
            [string] path of the excel file
        years:
            [list] list of the years you want to read
        months:
            [list] list of the months you you want to read

        Returns
        -------
         List of the values in the excel file

        examples
        --------
            years=[2009,2010,2011]#,2012,2013]
            months=[1,2,3,4,5,6,7,8,9,10,11,12]
            Q=ReadExcelData(path+"Discharge/Qout.xlsx",years,months)
        """

        Qout = pd.read_excel(path)
        Q = []
        #    years=[2009,2010,2011]#,2012,2013]
        #    months=[1,2,3,4,5,6,7,8,9,10,11,12]
        for year in years:
            for month in months:
                row = Qout[Qout["year"] == year][Qout["month"] == month]
                row = row.drop(["year", "month"], axis=1)
                row = row.values.tolist()[0]
                Q = Q + row

        Q = [Q[i] for i in range(len(Q)) if not np.isnan(Q[i])]

        return Q

    def ListAttributes(self):
        """Print Attributes List."""
        logger.info("\n")
        logger.info(
            f"Attributes List of: {repr(self.__dict__['name'])} - {self.__class__.__name__} Instance\n"
        )
        self_keys = list(self.__dict__.keys())
        self_keys.sort()
        for key in self_keys:
            if key != "name":
                logger.info(f"{key} : {repr(self.__dict__[key])}")

        logger.info("\n")



"""Created on Sat Apr 24 20:57:45 2021.

@author: mofarrag
"""

# import Hapi.hm.calibration as calibration
# import Hapi.hm.crosssection as crosssection
# import Hapi.hm.event as event
# import Hapi.hm.inputs as inputs
# import Hapi.hm.interface as interface
# import Hapi.hm.river as river
# import Hapi.hm.saintvenant as saintvenant
